@@ -39,6 +39,9 @@
 #ifdef HAVE_LINUX_BLKPG_H
 #include <linux/blkpg.h>
 #endif
+#ifdef HAVE_LIBBLKID_INTERNAL
+#include <blkid.h>
+#endif
 
 #include "gpt.h"
 
@@ -174,7 +177,7 @@ static int type_open = O_RDWR;
  * Raw disk label. For DOS-type partition tables the MBR,
  * with descriptions of the primary partitions.
  */
-unsigned char MBRbuffer[MAX_SECTOR_SIZE];
+unsigned char *MBRbuffer;
 
 /*
  * per partition table entry data
@@ -219,6 +222,12 @@ unsigned int	heads,
 	display_in_cyl_units = 1;
 
 unsigned long long total_number_of_sectors;	/* (!) 512-byte sectors */
+unsigned long grain = DEFAULT_SECTOR_SIZE,
+	      io_size = DEFAULT_SECTOR_SIZE,
+	      min_io_size = DEFAULT_SECTOR_SIZE,
+	      phy_sector_size = DEFAULT_SECTOR_SIZE,
+	      alignment_offset;
+int has_topology;
 
 #define dos_label (!sun_label && !sgi_label && !aix_label && !mac_label && !osf_label)
 int	sun_label = 0;			/* looking at sun disklabel */
@@ -233,6 +242,7 @@ jmp_buf listingbuf;
 void fatal(enum failure why) {
 	char	error[LINE_LENGTH],
 		*message = error;
+	int	rc = EXIT_FAILURE;
 
 	if (listing) {
 		close(fd);
@@ -240,25 +250,23 @@ void fatal(enum failure why) {
 	}
 
 	switch (why) {
+		case help:
+			rc = EXIT_SUCCESS;
 		case usage: message = _(
-"Usage: fdisk [-b SSZ] [-u] DISK     Change partition table\n"
-"       fdisk -l [-b SSZ] [-u] DISK  List partition table(s)\n"
-"       fdisk -s PARTITION           Give partition size(s) in blocks\n"
-"       fdisk -v                     Give fdisk version\n"
-"Here DISK is something like /dev/hdb or /dev/sda\n"
-"and PARTITION is something like /dev/hda7\n"
-"-u: give Start and End in sector (instead of cylinder) units\n"
-"-b 2048: (for certain MO disks) use 2048-byte sectors\n");
-			break;
-		case usage2:
-		  /* msg in cases where fdisk used to probe */
-			message = _(
-"Usage: fdisk [-l] [-b SSZ] [-u] device\n"
-"E.g.: fdisk /dev/hda  (for the first IDE disk)\n"
-"  or: fdisk /dev/sdc  (for the third SCSI disk)\n"
-"  or: fdisk /dev/eda  (for the first PS/2 ESDI drive)\n"
-"  or: fdisk /dev/rd/c0d0  or: fdisk /dev/ida/c0d0  (for RAID devices)\n"
-"  ...\n");
+"Usage:\n"
+" fdisk [options] <disk>    change partition table\n"
+" fdisk [options] -l <disk> list partition table(s)\n"
+" fdisk -s <partition>      give partition size(s) in blocks\n"
+"\nOptions:\n"
+" -b <size>                 sector size (512, 1024, 2048 or 4096)\n"
+" -c                        switch off DOS-compatible mode\n"
+" -h                        print help\n"
+" -u <size>                 give sizes in sectors instead of cylinders\n"
+" -v                        print version\n"
+" -C <number>               specify the number of cylinders\n"
+" -H <number>               specify the number of heads\n"
+" -S <number>               specify the number of sectors per track\n"
+"\n");
 			break;
 		case unable_to_open:
 			snprintf(error, sizeof(error),
@@ -290,7 +298,7 @@ void fatal(enum failure why) {
 
 	fputc('\n', stderr);
 	fputs(message, stderr);
-	exit(1);
+	exit(rc);
 }
 
 static void
@@ -384,7 +392,7 @@ static void
 menu(void) {
 	if (sun_label) {
 	   puts(_("Command action"));
-	   puts(_("   a   toggle a read only flag")); 		/* sun */
+	   puts(_("   a   toggle a read only flag"));		/* sun */
 	   puts(_("   b   edit bsd disklabel"));
 	   puts(_("   c   toggle the mountable flag"));		/* sun */
 	   puts(_("   d   delete a partition"));
@@ -633,6 +641,90 @@ test_c(char **m, char *mesg) {
 	return val;
 }
 
+#define alignment_required	(grain != sector_size)
+
+static int
+lba_is_aligned(unsigned long long lba)
+{
+	unsigned int granularity = max(phy_sector_size, min_io_size);
+	unsigned long long offset = (lba << 9) & (granularity - 1);
+
+	return !((granularity + alignment_offset - offset) & (granularity - 1));
+}
+
+#define ALIGN_UP	1
+#define ALIGN_DOWN	2
+#define ALIGN_NEAREST	3
+
+static unsigned long long
+align_lba(unsigned long long lba, int direction)
+{
+	unsigned long long res;
+
+	if (lba_is_aligned(lba))
+		res = lba;
+	else {
+		unsigned long long sects_in_phy = grain / sector_size;
+
+		if (lba < sector_offset)
+			res = sector_offset;
+
+		else if (direction == ALIGN_UP)
+			res = ((lba + sects_in_phy) / sects_in_phy) * sects_in_phy;
+
+		else if (direction == ALIGN_DOWN)
+			res = (lba / sects_in_phy) * sects_in_phy;
+
+		else /* ALIGN_NEAREST */
+			res = ((lba + sects_in_phy / 2) / sects_in_phy) * sects_in_phy;
+
+		if (alignment_offset && !lba_is_aligned(res) &&
+		    res > alignment_offset / sector_size) {
+			/*
+			 * apply alignment_offset
+			 *
+			 * On disk with alignment compensation physical blocks starts
+			 * at LBA < 0 (usually LBA -1). It means we have to move LBA
+			 * according the offset to be on the physical boundary.
+			 */
+			/* fprintf(stderr, "LBA: %llu apply alignment_offset\n", res); */
+			res -= (max(phy_sector_size, min_io_size) -
+					alignment_offset) / sector_size;
+
+			if (direction == ALIGN_UP && res < lba)
+				res += sects_in_phy;
+		}
+	}
+
+	/***
+	 fprintf(stderr, "LBA %llu (%s) --align-(%s)--> %llu (%s)\n",
+				lba,
+				lba_is_aligned(lba) ? "OK" : "FALSE",
+				direction == ALIGN_UP ?   "UP     " :
+				direction == ALIGN_DOWN ? "DOWN   " : "NEAREST",
+				res,
+				lba_is_aligned(res) ? "OK" : "FALSE");
+	***/
+	return res;
+}
+
+static unsigned long long
+align_lba_in_range(	unsigned long long lba,
+			unsigned long long start,
+			unsigned long long stop)
+{
+	start = align_lba(start, ALIGN_UP);
+	stop = align_lba(stop, ALIGN_DOWN);
+
+	lba = align_lba(lba, ALIGN_NEAREST);
+
+	if (lba < start)
+		return start;
+	else if (lba > stop)
+		return stop;
+	return lba;
+}
+
 static int
 warn_geometry(void) {
 	char *m = NULL;
@@ -665,18 +757,8 @@ void update_units(void)
 }
 
 static void
-warn_cylinders(void) {
-	if (dos_label && cylinders > 1024 && !nowarn)
-		fprintf(stderr, _("\n"
-"The number of cylinders for this disk is set to %d.\n"
-"There is nothing wrong with that, but this is larger than 1024,\n"
-"and could in certain setups cause problems with:\n"
-"1) software that runs at boot time (e.g., old versions of LILO)\n"
-"2) booting and partitioning software from other OSs\n"
-"   (e.g., DOS FDISK, OS/2 FDISK)\n"),
-			cylinders);
-
-	if (total_number_of_sectors > UINT_MAX) {
+warn_limits(void) {
+	if (total_number_of_sectors > UINT_MAX && !nowarn) {
 		int giga = (total_number_of_sectors << 9) / 1000000000;
 		int hectogiga = (giga + 50) / 100;
 
@@ -692,6 +774,35 @@ warn_cylinders(void) {
 	}
 }
 
+static void
+warn_alignment(void) {
+	if (nowarn)
+		return;
+
+	if (sector_size != phy_sector_size)
+		fprintf(stderr, _("\n"
+"The device presents a logical sector size that is smaller than\n"
+"the physical sector size. Aligning to a physical sector (or optimal\n"
+"I/O) size boundary is recommended, or performance may be impacted.\n"));
+
+	if (dos_compatible_flag) {
+		fprintf(stderr, _("\n"
+"WARNING: DOS-compatible mode is deprecated. It's strongly recommended to\n"
+"         switch off the mode (command 'c')"));
+
+		if (display_in_cyl_units)
+			fprintf(stderr, _(" and change display units to\n"
+"         sectors (command 'u').\n"));
+		else
+			fprintf(stderr, ".\n");
+
+	 } else if (display_in_cyl_units)
+		fprintf(stderr, _("\n"
+"WARNING: cylinders as display units are deprecated. Use command 'u' to\n"
+"         change units to sectors.\n"));
+
+}
+
 static void
 read_extended(int ext) {
 	int i;
@@ -849,14 +960,52 @@ create_doslabel(void) {
 }
 
 static void
-get_sectorsize(int fd) {
+get_topology(int fd) {
 	int arg;
+#ifdef HAVE_LIBBLKID_INTERNAL
+	blkid_probe pr;
 
 	if (user_set_sector_size)
 		return;
+	pr = blkid_new_probe();
+	if (pr && blkid_probe_set_device(pr, fd, 0, 0) == 0) {
+		blkid_topology tp = blkid_probe_get_topology(pr);
 
-	if (blkdev_get_sector_size(fd, &arg) == 0)
+		if (tp) {
+			min_io_size = blkid_topology_get_minimum_io_size(tp);
+			io_size = blkid_topology_get_optimal_io_size(tp);
+			phy_sector_size = blkid_topology_get_physical_sector_size(tp);
+			alignment_offset = blkid_topology_get_alignment_offset(tp);
+
+			/* We assume that the device provides topology info if
+			 * optimal_io_size is set or alignment_offset is set or
+			 * minimum_io_size is not power of 2.
+			 *
+			 * See also update_sector_offset().
+			 */
+			if (io_size || alignment_offset ||
+			    (min_io_size & (min_io_size - 1)))
+				has_topology = 1;
+			if (!io_size)
+				/* optimal IO is optional, default to minimum IO */
+				io_size = min_io_size;
+		}
+	}
+	blkid_free_probe(pr);
+#endif
+
+	if (user_set_sector_size)
+		;
+	else if (blkdev_get_sector_size(fd, &arg) == 0)
 		sector_size = arg;
+
+	if (!phy_sector_size)
+		phy_sector_size = sector_size;
+	if (!min_io_size)
+		min_io_size = phy_sector_size;
+	if (!io_size)
+		io_size = min_io_size;
+
 	if (sector_size != DEFAULT_SECTOR_SIZE)
 		printf(_("Note: sector size is %d (not %d)\n"),
 		       sector_size, DEFAULT_SECTOR_SIZE);
@@ -907,11 +1056,61 @@ get_partition_table_geometry(void) {
 	}
 }
 
+/*
+ * Sets LBA of the first partition
+ */
+void
+update_sector_offset(void)
+{
+	grain = io_size;
+
+	if (dos_compatible_flag)
+		sector_offset = sectors;	/* usually 63 sectors */
+	else {
+		/*
+		 * Align the begin of partitions to:
+		 *
+		 * a) topology
+		 *  a2) alignment offset
+		 *  a1) or physical sector (minimal_io_size, aka "grain")
+		 *
+		 * b) or default to 1MiB (2048 sectrors, Windows Vista default)
+		 *
+		 * c) or for very small devices use 1 phy.sector
+		 */
+		unsigned long long x = 0;
+
+		if (has_topology) {
+			if (alignment_offset)
+				x = alignment_offset;
+			else if (io_size > 2048 * 512)
+				x = io_size;
+		}
+		/* default to 1MiB */
+		if (!x)
+			x = 2048 * 512;
+
+		sector_offset = x / sector_size;
+
+		/* don't use huge offset on small devices */
+		if (total_number_of_sectors <= sector_offset * 4)
+			sector_offset = phy_sector_size / sector_size;
+
+		/* use 1MiB grain always when possible */
+		if (grain < 2048 * 512)
+			grain = 2048 * 512;
+
+		/* don't use huge grain on small devices */
+		if (total_number_of_sectors <= (grain * 4 / sector_size))
+			grain = phy_sector_size;
+	}
+}
+
 void
 get_geometry(int fd, struct geom *g) {
 	unsigned long long llcyls;
 
-	get_sectorsize(fd);
+	get_topology(fd);
 	sector_factor = sector_size / 512;
 	guess_device_type(fd);
 	heads = cylinders = sectors = 0;
@@ -931,9 +1130,7 @@ get_geometry(int fd, struct geom *g) {
 	if (blkdev_get_sectors(fd, &total_number_of_sectors) == -1)
 		total_number_of_sectors = 0;
 
-	sector_offset = 1;
-	if (dos_compatible_flag)
-		sector_offset = sectors;
+	update_sector_offset();
 
 	llcyls = total_number_of_sectors / (heads * sectors * sector_factor);
 	cylinders = llcyls;
@@ -949,6 +1146,27 @@ get_geometry(int fd, struct geom *g) {
 	}
 }
 
+/*
+ * Please, always use allocated buffer if you want to cast the buffer to
+ * any struct -- cast non-allocated buffer to any struct is against
+ * strict-aliasing rules.  --kzak 16-Oct-2009
+ */
+static void init_mbr_buffer(void)
+{
+	if (MBRbuffer)
+		return;
+
+	MBRbuffer = calloc(1, MAX_SECTOR_SIZE);
+	if (!MBRbuffer)
+		fatal(out_of_memory);
+}
+
+void zeroize_mbr_buffer(void)
+{
+	if (MBRbuffer)
+		memset(MBRbuffer, 0, MAX_SECTOR_SIZE);
+}
+
 /*
  * Read MBR.  Returns:
  *   -1: no 0xaa55 flag present (possibly entire disk BSD)
@@ -1051,9 +1269,6 @@ got_dos_table:
 		}
 	}
 
-	warn_cylinders();
-	warn_geometry();
-
 	for (i = 0; i < 4; i++) {
 		struct pte *pe = &ptes[i];
 
@@ -1078,6 +1293,10 @@ got_dos_table:
 		}
 	}
 
+	warn_geometry();
+	warn_limits();
+	warn_alignment();
+
 	return 0;
 }
 
@@ -1146,16 +1365,9 @@ read_hex(struct systypes *sys)
         }
 }
 
-/*
- * Print the message MESG, then read an integer in LOW..HIGH.
- * If the user hits Enter, DFLT is returned, provided that is in LOW..HIGH.
- * Answers like +10 are interpreted as offsets from BASE.
- *
- * There is no default if DFLT is not between LOW and HIGH.
- */
-unsigned int
-read_int(unsigned int low, unsigned int dflt, unsigned int high,
-	 unsigned int base, char *mesg)
+static unsigned int
+read_int_sx(unsigned int low, unsigned int dflt, unsigned int high,
+	 unsigned int base, char *mesg, int *suffix)
 {
 	unsigned int i;
 	int default_ok = 1;
@@ -1254,6 +1466,8 @@ read_int(unsigned int low, unsigned int dflt, unsigned int high,
 				bytes += unit/2;	/* round */
 				bytes /= unit;
 				i = bytes;
+				if (suffix)
+					*suffix = absolute;
 			}
 			if (minus)
 				i = -i;
@@ -1275,6 +1489,21 @@ read_int(unsigned int low, unsigned int dflt, unsigned int high,
 	return i;
 }
 
+/*
+ * Print the message MESG, then read an integer in LOW..HIGH.
+ * If the user hits Enter, DFLT is returned, provided that is in LOW..HIGH.
+ * Answers like +10 are interpreted as offsets from BASE.
+ *
+ * There is no default if DFLT is not between LOW and HIGH.
+ */
+unsigned int
+read_int(unsigned int low, unsigned int dflt, unsigned int high,
+	 unsigned int base, char *mesg)
+{
+	return read_int_sx(low, dflt, high, base, mesg, NULL);
+}
+
+
 int
 get_partition(int warn, int max) {
 	struct pte *pe;
@@ -1381,14 +1610,12 @@ toggle_active(int i) {
 static void
 toggle_dos_compatibility_flag(void) {
 	dos_compatible_flag = ~dos_compatible_flag;
-	if (dos_compatible_flag) {
-		sector_offset = sectors;
+	if (dos_compatible_flag)
 		printf(_("DOS Compatibility flag is set\n"));
-	}
-	else {
-		sector_offset = 1;
+	else
 		printf(_("DOS Compatibility flag is not set\n"));
-	}
+
+	update_sector_offset();
 }
 
 static void
@@ -1567,6 +1794,9 @@ static void check_consistency(struct partition *p, int partition) {
 	unsigned int lbc, lbh, lbs;	/* logical beginning c, h, s */
 	unsigned int lec, leh, les;	/* logical ending c, h, s */
 
+	if (!dos_compatible_flag)
+		return;
+
 	if (!heads || !sectors || (partition >= 4))
 		return;		/* do not check extended partitions */
 
@@ -1624,6 +1854,14 @@ static void check_consistency(struct partition *p, int partition) {
 	}
 }
 
+static void
+check_alignment(unsigned long long lba, int partition)
+{
+	if (!lba_is_aligned(lba))
+		printf(_("Partition %i does not start on physical sector boundary.\n"),
+			partition + 1);
+}
+
 static void
 list_disk_geometry(void) {
 	long long bytes = (total_number_of_sectors << 9);
@@ -1646,6 +1884,13 @@ list_disk_geometry(void) {
 	printf(_("Units = %s of %d * %d = %d bytes\n"),
 	       str_units(PLURAL),
 	       units_per_sector, sector_size, units_per_sector * sector_size);
+
+	printf(_("Sector size (logical/physical): %u bytes / %lu bytes\n"),
+				sector_size, phy_sector_size);
+	printf(_("I/O size (minimum/optimal): %lu bytes / %lu bytes\n"),
+				min_io_size, io_size);
+	if (alignment_offset)
+		printf(_("Alignment offset: %lu bytes\n"), alignment_offset);
 	if (dos_label)
 		dos_print_mbr_id();
 	printf("\n");
@@ -1854,6 +2099,7 @@ list_table(int xtra) {
 /* type name */		(type = partition_type(p->sys_ind)) ?
 			type : _("Unknown"));
 			check_consistency(p, i);
+			check_alignment(get_partition_start(pe), i);
 		}
 	}
 
@@ -1886,8 +2132,10 @@ x_list_table(int extend) {
 				cylinder(p->end_sector, p->end_cyl),
 				(unsigned long) get_start_sect(p),
 				(unsigned long) get_nr_sects(p), p->sys_ind);
-			if (p->sys_ind)
+			if (p->sys_ind) {
 				check_consistency(p, i);
+				check_alignment(get_partition_start(pe), i);
+			}
 		}
 	}
 }
@@ -1964,6 +2212,7 @@ verify(void) {
 		p = pe->part_table;
 		if (p->sys_ind && !IS_EXTENDED (p->sys_ind)) {
 			check_consistency(p, i);
+			check_alignment(get_partition_start(pe), i);
 			if (get_partition_start(pe) < first[i])
 				printf(_("Warning: bad start-of-data in "
 					"partition %d\n"), i + 1);
@@ -2007,10 +2256,31 @@ verify(void) {
 		printf(_("Total allocated sectors %llu greater than the maximum"
 			" %llu\n"), total, n_sectors);
 	else if (total < n_sectors)
-		printf(_("%lld unallocated %d-byte sectors\n"),
+		printf(_("Remaining %lld unallocated %d-byte sectors\n"),
 		       n_sectors - total, sector_size);
 }
 
+static unsigned long long
+get_unused_start(int part_n,
+		unsigned long long start,
+		unsigned long long first[],
+		unsigned long long last[])
+{
+	int i;
+
+	for (i = 0; i < partitions; i++) {
+		unsigned long long lastplusoff;
+
+		if (start == ptes[i].offset)
+			start += sector_offset;
+		lastplusoff = last[i] + ((part_n < 4) ? 0 : sector_offset);
+		if (start >= first[i] && start <= lastplusoff)
+			start = lastplusoff + 1;
+	}
+
+	return start;
+}
+
 static void
 add_partition(int n, int sys) {
 	char mesg[256];		/* 48 does not suffice in Japanese */
@@ -2051,16 +2321,19 @@ add_partition(int n, int sys) {
 
 	snprintf(mesg, sizeof(mesg), _("First %s"), str_units(SINGULAR));
 	do {
-		temp = start;
-		for (i = 0; i < partitions; i++) {
-			unsigned long long lastplusoff;
+		unsigned long long dflt, aligned;
 
-			if (start == ptes[i].offset)
-				start += sector_offset;
-			lastplusoff = last[i] + ((n<4) ? 0 : sector_offset);
-			if (start >= first[i] && start <= lastplusoff)
-				start = lastplusoff + 1;
-		}
+		temp = start;
+		dflt = start = get_unused_start(n, start, first, last);
+
+		/* the default sector should be aligned and unused */
+		do {
+			aligned = align_lba_in_range(dflt, dflt, limit);
+			dflt = get_unused_start(n, aligned, first, last);
+		} while (dflt != aligned && dflt > aligned && dflt < limit);
+
+		if (dflt >= limit)
+			dflt = start;
 		if (start > limit)
 			break;
 		if (start >= temp+units_per_sector && read) {
@@ -2071,7 +2344,7 @@ add_partition(int n, int sys) {
 		if (!read && start == temp) {
 			unsigned long long i = start;
 
-			start = read_int(cround(i), cround(i), cround(limit),
+			start = read_int(cround(i), cround(dflt), cround(limit),
 					 0, mesg);
 			if (display_in_cyl_units) {
 				start = (start - 1) * units_per_sector;
@@ -2108,17 +2381,30 @@ add_partition(int n, int sys) {
 	if (cround(start) == cround(limit)) {
 		stop = limit;
 	} else {
+		int sx = 0;
+
 		snprintf(mesg, sizeof(mesg),
 			_("Last %1$s, +%2$s or +size{K,M,G}"),
 			 str_units(SINGULAR), str_units(PLURAL));
 
-		stop = read_int(cround(start), cround(limit), cround(limit),
-				cround(start), mesg);
+		stop = read_int_sx(cround(start), cround(limit), cround(limit),
+				cround(start), mesg, &sx);
 		if (display_in_cyl_units) {
 			stop = stop * units_per_sector - 1;
 			if (stop >limit)
 				stop = limit;
 		}
+
+		if (sx && alignment_required) {
+			/* the last sector has not been exactly requested (but
+			 * defined by +size{K,M,G} convention), so be smart
+			 * and align the end of the partition. The next
+			 * partition will start at phy.block boundary.
+			 */
+			stop = align_lba_in_range(stop, start, limit) - 1;
+			if (stop > limit)
+				stop = limit;
+		}
 	}
 
 	set_partition(n, 0, start, stop, sys);
@@ -2282,7 +2568,6 @@ reread_partition_table(int leave) {
 	i = fstat(fd, &statbuf);
 	if (i == 0 && S_ISBLK(statbuf.st_mode)) {
 		sync();
-		sleep(2);
 #ifdef BLKRRPART
 		printf(_("Calling ioctl() to re-read partition table.\n"));
 		i = ioctl(fd, BLKRRPART);
@@ -2313,7 +2598,6 @@ reread_partition_table(int leave) {
 
 		printf(_("Syncing disks.\n"));
 		sync();
-		sleep(4);		/* for sync() */
 		exit(!!i);
 	}
 }
@@ -2395,8 +2679,6 @@ xselect(void) {
 					 _("Number of cylinders"));
 			if (sun_label)
 				sun_set_ncyl(cylinders);
-			if (dos_label)
-				warn_cylinders();
 			break;
 		case 'd':
 			print_raw();
@@ -2447,12 +2729,11 @@ xselect(void) {
 		case 's':
 			user_sectors = sectors = read_int(1, sectors, 63, 0,
 					   _("Number of sectors"));
-			if (dos_compatible_flag) {
-				sector_offset = sectors;
+			if (dos_compatible_flag)
 				fprintf(stderr, _("Warning: setting "
 					"sector offset for DOS "
 					"compatiblity\n"));
-			}
+			update_sector_offset();
 			update_units();
 			break;
 		case 'v':
@@ -2594,17 +2875,7 @@ main(int argc, char **argv) {
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
 
-	/*
-	 * Calls:
-	 *  fdisk -v
-	 *  fdisk -l [-b sectorsize] [-u] device ...
-	 *  fdisk -s [partition] ...
-	 *  fdisk [-b sectorsize] [-u] device
-	 *
-	 * Options -C, -H, -S set the geometry.
-	 *
-	 */
-	while ((c = getopt(argc, argv, "b:C:H:lsS:uvV")) != -1) {
+	while ((c = getopt(argc, argv, "b:cC:hH:lsS:uvV")) != -1) {
 		switch (c) {
 		case 'b':
 			/* Ugly: this sector size is really per device,
@@ -2621,6 +2892,12 @@ main(int argc, char **argv) {
 		case 'C':
 			user_cylinders = atoi(optarg);
 			break;
+		case 'c':
+			dos_compatible_flag = 0;
+			break;
+		case 'h':
+			fatal(help);
+			break;
 		case 'H':
 			user_heads = atoi(optarg);
 			if (user_heads <= 0 || user_heads > 256)
@@ -2658,6 +2935,8 @@ main(int argc, char **argv) {
 			 " be used with one specified device\n"));
 #endif
 
+	init_mbr_buffer();
+
 	if (optl) {
 		nowarn = 1;
 		type_open = O_RDONLY;
@@ -2704,10 +2983,8 @@ main(int argc, char **argv) {
 
 	if (argc-optind == 1)
 		disk_device = argv[optind];
-	else if (argc-optind != 0)
-		fatal(usage);
 	else
-		fatal(usage2);
+		fatal(usage);
 
 	gpt_warning(disk_device);
 	get_boot(fdisk);
