@@ -1,6 +1,5 @@
 /*
- * A mount(8) for Linux 0.99.
- * mount.c,v 1.1.1.1 1993/11/18 08:40:51 jrs Exp
+ * A mount(8) for Linux.
  *
  * Modifications by many people. Distributed under GPL.
  */
@@ -30,6 +29,7 @@
 
 #include "pathnames.h"
 #include "fsprobe.h"
+#include "devname.h"
 #include "mount_constants.h"
 #include "sundries.h"
 #include "xmalloc.h"
@@ -41,7 +41,6 @@
 #include "env.h"
 #include "nls.h"
 #include "blkdev.h"
-#include "realpath.h"
 
 #define DO_PS_FIDDLING
 
@@ -55,7 +54,7 @@ static int fake = 0;
 /* True if we are allowed to call /sbin/mount.${FSTYPE} */
 static int external_allowed = 1;
 
-/* Don't write a entry in /etc/mtab (-n).  */
+/* Don't write an entry in /etc/mtab (-n).  */
 static int nomtab = 0;
 
 /* True for explicit readonly (-r).  */
@@ -81,12 +80,21 @@ static int list_with_volumelabel = 0;
  */
 static int mounttype = 0;
 
-/* True if ruid != euid.  */
-static int suid = 0;
+/* True if (ruid != euid) or (0 != ruid), i.e. only "user" mounts permitted.  */
+static int restricted = 1;
 
 /* Contains the fd to read the passphrase from, if any. */
 static int pfd = -1;
 
+/* mount(2) options */
+struct mountargs {
+       const char *spec;
+       const char *node;
+       const char *type;
+       int flags;
+       void *data;
+};
+
 /* Map from -o and fstab option strings to the flag argument to mount(2).  */
 struct opt_map {
   const char *opt;		/* option name */
@@ -166,6 +174,10 @@ static const struct opt_map opt_map[] = {
   { "atime",	0, 1, MS_NOATIME },	/* Update access time */
   { "noatime",	0, 0, MS_NOATIME },	/* Do not update access time */
 #endif
+#ifdef MS_I_VERSION
+  { "iversion",	0, 0, MS_I_VERSION },	/* Update inode I_version time */
+  { "noiversion", 0, 1, MS_I_VERSION },	/* Don't update inode I_version time */
+#endif
 #ifdef MS_NODIRATIME
   { "diratime",	0, 1, MS_NODIRATIME },	/* Update dir access times */
   { "nodiratime", 0, 0, MS_NODIRATIME },/* Do not update dir access times */
@@ -175,6 +187,10 @@ static const struct opt_map opt_map[] = {
 					  mtime/ctime */
   { "norelatime", 0, 1, MS_RELATIME }, /* Update access time without regard
 					  to mtime/ctime */
+#endif
+#ifdef MS_STRICTATIME
+  { "strictatime", 0, 0, MS_STRICTATIME }, /* Strict atime semantics */
+  { "nostrictatime", 0, 1, MS_STRICTATIME }, /* kernel default atime */
 #endif
   { "nofail",	0, 0, MS_COMMENT},	/* Do not fail if ENOENT on dev */
   { NULL,	0, 0, 0		}
@@ -185,6 +201,7 @@ static int opt_nofail = 0;
 static const char *opt_loopdev, *opt_vfstype, *opt_offset, *opt_sizelimit,
         *opt_encryption, *opt_speed, *opt_comment, *opt_uhelper;
 
+static int is_readonly(const char *node);
 static int mounted (const char *spec0, const char *node0);
 static int check_special_mountprog(const char *spec, const char *node,
 		const char *type, int flags, char *extra_opts, int *status);
@@ -238,8 +255,8 @@ print_one (const struct my_mntent *me) {
 		printf (" type %s", me->mnt_type);
 	if (me->mnt_opts != NULL)
 		printf (" (%s)", me->mnt_opts);
-	if (list_with_volumelabel) {
-		const char *devname = fsprobe_get_devname(me->mnt_fsname);
+	if (list_with_volumelabel && is_pseudo_fs(me->mnt_type) == 0) {
+		const char *devname = spec_to_devname(me->mnt_fsname);
 
 		if (devname) {
 			const char *label;
@@ -266,6 +283,14 @@ print_all (char *types) {
 	  if (matching_type (mc->m.mnt_type, types))
 	       print_one (&(mc->m));
      }
+
+     if (!mtab_does_not_exist() && !mtab_is_a_symlink() && is_readonly(_PATH_MOUNTED))
+          printf(_("\n"
+	"mount: warning: /etc/mtab is not writable (e.g. read-only filesystem).\n"
+	"       It's possible that information reported by mount(8) is not\n"
+	"       up to date. For actual information about system mount points\n"
+	"       check the /proc/mounts file.\n\n"));
+
      exit (0);
 }
 
@@ -327,7 +352,7 @@ append_context(const char *optname, char *optdata, char **extra_opts)
 	security_context_t raw = NULL;
 	char *data = NULL;
 
-	if (!is_selinux_enabled())
+	if (is_selinux_enabled() != 1)
 		/* ignore the option if we running without selinux */
 		return 0;
 
@@ -338,8 +363,8 @@ append_context(const char *optname, char *optdata, char **extra_opts)
 	data = *optdata =='"' ? strip_quotes(optdata) : optdata;
 
 	if (selinux_trans_to_raw_context(
-			(security_context_t) data, &raw)==-1 ||
-			raw==NULL)
+			(security_context_t) data, &raw) == -1 ||
+			raw == NULL)
 		return -1;
 
 	if (verbose)
@@ -419,6 +444,10 @@ parse_opt(char *opt, int *mask, char **extra_opts) {
 		if (append_context("defcontext=", opt+11, extra_opts) == 0)
 			return;
 	}
+	if (strncmp(opt, "rootcontext=", 12) == 0 && *(opt+12)) {
+		if (append_context("rootcontext=", opt+12, extra_opts) == 0)
+			return;
+	}
 #endif
 	*extra_opts = append_opt(*extra_opts, opt, NULL);
 }
@@ -496,9 +525,11 @@ fix_opts_string (int flags, const char *extra_opts, const char *user) {
 }
 
 static int
-already (const char *spec, const char *node) {
+already (const char *spec0, const char *node0) {
 	struct mntentchn *mc;
 	int ret = 1;
+	char *spec = canonicalize_spec(spec0);
+	char *node = canonicalize(node0);
 
 	if ((mc = getmntfile(node)) != NULL)
 		error (_("mount: according to mtab, "
@@ -510,6 +541,10 @@ already (const char *spec, const char *node) {
 		       spec, mc->m.mnt_dir);
 	else
 		ret = 0;
+
+	free(spec);
+	free(node);
+
 	return ret;
 }
 
@@ -535,7 +570,7 @@ create_mtab (void) {
 		char *extra_opts;
 		parse_opts (fstab->m.mnt_opts, &flags, &extra_opts);
 		mnt.mnt_dir = "/";
-		mnt.mnt_fsname = fsprobe_get_devname(fstab->m.mnt_fsname);
+		mnt.mnt_fsname = spec_to_devname(fstab->m.mnt_fsname);
 		mnt.mnt_type = fstab->m.mnt_type;
 		mnt.mnt_opts = fix_opts_string (flags, extra_opts, NULL);
 		mnt.mnt_freq = mnt.mnt_passno = 0;
@@ -684,6 +719,157 @@ check_special_mountprog(const char *spec, const char *node, const char *type, in
 }
 
 
+/* list of already tested filesystems by procfsloop_mount() */
+static struct tried {
+	struct tried *next;
+	char *type;
+} *tried = NULL;
+
+static int
+was_tested(const char *fstype) {
+	struct tried *t;
+
+	if (fsprobe_known_fstype(fstype))
+		return 1;
+	for (t = tried; t; t = t->next) {
+		if (!strcmp(t->type, fstype))
+			return 1;
+	}
+	return 0;
+}
+
+static void
+set_tested(const char *fstype) {
+	struct tried *t = xmalloc(sizeof(struct tried));
+
+	t->next = tried;
+	t->type = xstrdup(fstype);
+	tried = t;
+}
+
+static void
+free_tested(void) {
+	struct tried *t, *tt;
+
+	t = tried;
+	while(t) {
+		free(t->type);
+		tt = t->next;
+		free(t);
+		t = tt;
+	}
+	tried = NULL;
+}
+
+static char *
+procfsnext(FILE *procfs) {
+   char line[100];
+   char fsname[100];
+
+   while (fgets(line, sizeof(line), procfs)) {
+      if (sscanf (line, "nodev %[^\n]\n", fsname) == 1) continue;
+      if (sscanf (line, " %[^ \n]\n", fsname) != 1) continue;
+      return xstrdup(fsname);
+   }
+   return 0;
+}
+
+/* Only use /proc/filesystems here, this is meant to test what
+   the kernel knows about, so /etc/filesystems is irrelevant.
+   Return: 1: yes, 0: no, -1: cannot open procfs */
+static int
+known_fstype_in_procfs(const char *type)
+{
+    FILE *procfs;
+    char *fsname;
+    int ret = -1;
+
+    procfs = fopen(_PATH_PROC_FILESYSTEMS, "r");
+    if (procfs) {
+	ret = 0;
+	while ((fsname = procfsnext(procfs)) != NULL)
+	    if (!strcmp(fsname, type)) {
+		ret = 1;
+		break;
+	    }
+	fclose(procfs);
+	procfs = NULL;
+    }
+    return ret;
+}
+
+/* Try all types in FILESYSTEMS, except those in *types,
+   in case *types starts with "no" */
+/* return: 0: OK, -1: error in errno, 1: type not found */
+/* when 0 or -1 is returned, *types contains the type used */
+/* when 1 is returned, *types is NULL */
+static int
+procfsloop_mount(int (*mount_fn)(struct mountargs *, int *, int *),
+			 struct mountargs *args,
+			 const char **types,
+			 int *special, int *status)
+{
+	char *files[2] = { _PATH_FILESYSTEMS, _PATH_PROC_FILESYSTEMS };
+	FILE *procfs;
+	char *fsname;
+	const char *notypes = NULL;
+	int no = 0;
+	int ret = 1;
+	int errsv = 0;
+	int i;
+
+	if (*types && !strncmp(*types, "no", 2)) {
+		no = 1;
+		notypes = (*types) + 2;
+	}
+	*types = NULL;
+
+	/* Use _PATH_PROC_FILESYSTEMS only when _PATH_FILESYSTEMS
+	 * (/etc/filesystems) does not exist.  In some cases trying a
+	 * filesystem that the kernel knows about on the wrong data will crash
+	 * the kernel; in such cases _PATH_FILESYSTEMS can be used to list the
+	 * filesystems that we are allowed to try, and in the order they should
+	 * be tried.  End _PATH_FILESYSTEMS with a line containing a single '*'
+	 * only, if _PATH_PROC_FILESYSTEMS should be tried afterwards.
+	 */
+	for (i=0; i<2; i++) {
+		procfs = fopen(files[i], "r");
+		if (!procfs)
+			continue;
+		while ((fsname = procfsnext(procfs)) != NULL) {
+			if (!strcmp(fsname, "*")) {
+				fclose(procfs);
+				goto nexti;
+			}
+			if (was_tested (fsname))
+				continue;
+			if (no && matching_type(fsname, notypes))
+				continue;
+			set_tested (fsname);
+			args->type = fsname;
+			if (verbose)
+				printf(_("Trying %s\n"), fsname);
+			if ((*mount_fn) (args, special, status) == 0) {
+				*types = fsname;
+				ret = 0;
+				break;
+			} else if (errno != EINVAL &&
+				   known_fstype_in_procfs(fsname) == 1) {
+				*types = "guess";
+				ret = -1;
+				errsv = errno;
+				break;
+			}
+		}
+		free_tested();
+		fclose(procfs);
+		errno = errsv;
+		return ret;
+	nexti:;
+	}
+	return 1;
+}
+
 static const char *
 guess_fstype_by_devname(const char *devname)
 {
@@ -718,9 +904,6 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
    if (*types && strcasecmp (*types, "auto") == 0)
       *types = NULL;
 
-   if (flags & (MS_BIND | MS_MOVE))
-      *types = "none";
-
    if (!*types && !(flags & MS_REMOUNT)) {
       *types = guess_fstype_by_devname(spec);
       if (*types) {
@@ -737,11 +920,11 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
 
    /* Accept a comma-separated list of types, and try them one by one */
    /* A list like "nonfs,.." indicates types not to use */
-   if (*types && strncmp(*types, "no", 2) && index(*types,',')) {
+   if (*types && strncmp(*types, "no", 2) && strchr(*types,',')) {
       char *t = strdup(*types);
       char *p;
 
-      while((p = index(t,',')) != NULL) {
+      while((p = strchr(t,',')) != NULL) {
 	 *p = 0;
 	 args.type = *types = t;
 	 if (do_mount (&args, special, status) == 0)
@@ -757,16 +940,16 @@ guess_fstype_and_mount(const char *spec, const char *node, const char **types,
       return do_mount (&args, special, status);
    }
 
-   return fsprobe_procfsloop_mount(do_mount, &args, types, special, status);
+   return procfsloop_mount(do_mount, &args, types, special, status);
 }
 
 /*
- * suid_check()
+ * restricted_check()
  *	Die if the user is not allowed to do this.
  */
 static void
-suid_check(const char *spec, const char *node, int *flags, char **user) {
-  if (suid) {
+restricted_check(const char *spec, const char *node, int *flags, char **user) {
+  if (restricted) {
       /*
        * MS_OWNER: Allow owners to mount when fstab contains
        * the owner option.  Note that this should never be used
@@ -837,7 +1020,7 @@ is_mounted_same_loopfile(const char *node0, const char *loopfile, unsigned long
 	char *node;
 	int res = 0;
 
-	node = canonicalize_mountpoint(node0);
+	node = canonicalize(node0);
 
 	/* Search for mountpoint node in mtab,
 	 * procceed if any of these has the loop option set or
@@ -977,8 +1160,8 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 		  const char *opts, int flags, int freq, int pass) {
 	struct my_mntent mnt;
 
-	mnt.mnt_fsname = canonicalize (spec);
-	mnt.mnt_dir = canonicalize_mountpoint (node);
+	mnt.mnt_fsname = is_pseudo_fs(type) ? xstrdup(spec) : canonicalize(spec);
+	mnt.mnt_dir = canonicalize (node);
 	mnt.mnt_type = type;
 	mnt.mnt_opts = opts;
 	mnt.mnt_freq = freq;
@@ -992,7 +1175,7 @@ update_mtab_entry(const char *spec, const char *node, const char *type,
 	if (!nomtab && mtab_does_not_exist()) {
 		if (verbose > 1)
 			printf(_("mount: no %s found - creating it..\n"),
-			       MOUNTED);
+			       _PATH_MOUNTED);
 		create_mtab ();
 	}
 
@@ -1051,6 +1234,41 @@ cdrom_setspeed(const char *spec) {
 	}
 }
 
+/*
+ * Check if @node is read-only filesystem by access() or futimens().
+ *
+ * Note that access(2) uses real-UID (= useless for suid programs)
+ * and euidaccess(2) does not check for read-only FS.
+ */
+static int
+is_readonly(const char *node)
+{
+	int res = 0;
+
+	if (getuid() == geteuid()) {
+		if (access(node, W_OK) == -1 && errno == EROFS)
+			res = 1;
+	}
+#ifdef HAVE_FUTIMENS
+	else {
+		struct timespec times[2];
+		int fd = open(node, O_RDONLY);
+
+		if (fd < 0)
+			goto done;
+
+		times[0].tv_nsec = UTIME_NOW;	/* atime */
+		times[1].tv_nsec = UTIME_OMIT;	/* mtime */
+
+		if (futimens(fd, times) == -1 && errno == EROFS)
+			res = 1;
+		close(fd);
+	}
+done:
+#endif
+	return res;
+}
+
 /*
  * try_mount_one()
  *	Try to mount one file system.
@@ -1072,6 +1290,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   int loop = 0;
   const char *loopdev = 0, *loopfile = 0;
   struct stat statbuf;
+  int retries = 0;	/* Nr of retries for mount in case of ENOMEDIUM */
 
   /* copies for freeing on exit */
   const char *opts1, *spec1, *node1, *types1, *extra_opts1;
@@ -1095,7 +1314,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   if (mount_all && (flags & MS_NOAUTO))
       goto out;
 
-  suid_check(spec, node, &flags, &user);
+  restricted_check(spec, node, &flags, &user);
 
   /* The "mount -f" checks for for existing record in /etc/mtab (with
    * regular non-fake mount this is usually done by kernel)
@@ -1124,6 +1343,9 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   if (loop)
       opt_loopdev = loopdev;
 
+  if (flags & (MS_BIND | MS_MOVE | MS_PROPAGATION))
+      types = "none";
+
   /*
    * Call mount.TYPE for types that require a separate mount program.
    * For the moment these types are ncpfs and smbfs. Maybe also vxfs.
@@ -1134,6 +1356,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       goto out;
   }
 
+mount_retry:
   block_signals (SIG_BLOCK);
 
   if (!fake) {
@@ -1147,6 +1370,17 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
     }
   }
 
+  /* Kernel allows to use MS_RDONLY for bind mounts, but the read-only request
+   * could be silently ignored. Check it to avoid 'ro' in ntab and 'rw' in
+   * /proc/mounts.
+   */
+  if (!fake && mnt5_res == 0 &&
+      (flags & MS_BIND) && (flags & MS_RDONLY) && !is_readonly(node)) {
+
+      printf(_("mount: warning: %s seems to be mounted read-write.\n"), node);
+      flags &= ~MS_RDONLY;
+  }
+
   if (fake || mnt5_res == 0) {
       /* Mount succeeded, report this (if verbose) and write mtab entry.  */
 
@@ -1175,7 +1409,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   /* Mount failed, complain, but don't die.  */
 
   if (types == 0) {
-    if (suid)
+    if (restricted)
       error (_("mount: I could not determine the filesystem type, "
 	       "and none was specified"));
     else
@@ -1282,7 +1516,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       error (_("mount: %s: can't read superblock"), spec); break;
     case ENODEV:
     {
-      int pfs = fsprobe_known_fstype_in_procfs(types);
+      int pfs = known_fstype_in_procfs(types);
 
       if (pfs == 1 || !strcmp(types, "guess"))
         error(_("mount: %s: unknown device"), spec);
@@ -1301,13 +1535,13 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	    u++;
 	  }
 	}
-	if (u && fsprobe_known_fstype_in_procfs(lowtype) == 1)
+	if (u && known_fstype_in_procfs(lowtype) == 1)
 	  error (_("mount: probably you meant %s"), lowtype);
 	else if (!strncmp(lowtype, "iso", 3) &&
-			fsprobe_known_fstype_in_procfs("iso9660") == 1)
+			known_fstype_in_procfs("iso9660") == 1)
 	  error (_("mount: maybe you meant 'iso9660'?"));
 	else if (!strncmp(lowtype, "fat", 3) &&
-			fsprobe_known_fstype_in_procfs("vfat") == 1)
+			known_fstype_in_procfs("vfat") == 1)
 	  error (_("mount: maybe you meant 'vfat'?"));
 	free(lowtype);
       } else
@@ -1322,7 +1556,7 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	error (_("mount: %s is not a block device, and stat fails?"), spec);
       else if (S_ISBLK(statbuf.st_mode))
         error (_("mount: the kernel does not recognize %s as a block device\n"
-	       "       (maybe `insmod driver'?)"), spec);
+	       "       (maybe `modprobe driver'?)"), spec);
       else if (S_ISREG(statbuf.st_mode))
 	error (_("mount: %s is not a block device (maybe try `-o loop'?)"),
 		 spec);
@@ -1344,6 +1578,10 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
 	  error (_("mount: %s%s is write-protected but explicit `-w' flag given"),
 		 bd, spec);
 	  break;
+      } else if (flags & MS_REMOUNT) {
+	  error (_("mount: cannot remount %s%s read-write, is write-protected"),
+		 bd, spec);
+	  break;
       } else {
 	 opts = opts0;
 	 types = types0;
@@ -1363,6 +1601,17 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
       }
       break;
     }
+    case ENOMEDIUM:
+      if (retries < CRDOM_NOMEDIUM_RETRIES) {
+	      if (verbose)
+		      printf(_("mount: no medium found on %s ...trying again\n"),
+				 spec);
+              sleep(3);
+	      ++retries;
+              goto mount_retry;
+      }
+      error(_("mount: no medium found on %s"), spec);
+      break;
     default:
       error ("mount: %s", strerror (mnt_err)); break;
     }
@@ -1370,6 +1619,27 @@ try_mount_one (const char *spec0, const char *node0, const char *types0,
   res = EX_FAIL;
 
  out:
+
+#if defined(HAVE_LIBSELINUX) && defined(HAVE_SECURITY_GET_INITIAL_CONTEXT)
+  if (res != EX_FAIL && verbose && is_selinux_enabled() > 0) {
+      security_context_t raw = NULL, def = NULL;
+
+      if (getfilecon(node, &raw) > 0 &&
+		     security_get_initial_context("file", &def) == 0) {
+
+	  if (!selinux_file_context_cmp(raw, def))
+	      printf(_("mount: %s does not contain SELinux labels.\n"
+                   "       You just mounted an file system that supports labels which does not\n"
+                   "       contain labels, onto an SELinux box. It is likely that confined\n"
+                   "       applications will generate AVC messages and not be allowed access to\n"
+                   "       this file system.  For more details see restorecon(8) and mount(8).\n"),
+                   node);
+      }
+      freecon(raw);
+      freecon(def);
+  }
+#endif
+
   my_free(extra_opts1);
   my_free(spec1);
   my_free(node1);
@@ -1453,7 +1723,7 @@ mount_one (const char *spec, const char *node, const char *types,
 	if (types == NULL || (strncmp(types, "nfs", 3) &&
 			      strncmp(types, "cifs", 4) &&
 			      strncmp(types, "smbfs", 5))) {
-		nspec = fsprobe_get_devname_for_mounting(spec);
+		nspec = spec_to_devname(spec);
 		if (nspec)
 			spec = nspec;
 	}
@@ -1469,11 +1739,11 @@ mounted (const char *spec0, const char *node0) {
 	int ret = 0;
 
 	/* Handle possible UUID= and LABEL= in spec */
-	spec = fsprobe_get_devname(spec0);
+	spec = spec_to_devname(spec0);
 	if (!spec)
 		return ret;
 
-	node = canonicalize_mountpoint(node0);
+	node = canonicalize(node0);
 
 	mc0 = mtab_head();
 	for (mc = mc0->nxt; mc && mc != mc0; mc = mc->nxt)
@@ -1651,10 +1921,10 @@ static struct option longopts[] = {
 	{ "test-opts", 1, 0, 'O' },
 	{ "pass-fd", 1, 0, 'p' },
 	{ "types", 1, 0, 't' },
-	{ "bind", 0, 0, 128 },
-	{ "move", 0, 0, 133 },
+	{ "bind", 0, 0, 'B' },
+	{ "move", 0, 0, 'M' },
 	{ "guess-fstype", 1, 0, 134 },
-	{ "rbind", 0, 0, 135 },
+	{ "rbind", 0, 0, 'R' },
 	{ "make-shared", 0, 0, 136 },
 	{ "make-slave", 0, 0, 137 },
 	{ "make-private", 0, 0, 138 },
@@ -1663,6 +1933,7 @@ static struct option longopts[] = {
 	{ "make-rslave", 0, 0, 141 },
 	{ "make-rprivate", 0, 0, 142 },
 	{ "make-runbindable", 0, 0, 143 },
+	{ "no-canonicalize", 0, 0, 144 },
 	{ "internal-only", 0, 0, 'i' },
 	{ NULL, 0, 0, 0 }
 };
@@ -1745,7 +2016,7 @@ getfs(const char *spec, const char *uuid, const char *label)
 	else if (label)
 		devname = fsprobe_get_devname_by_label(label);
 	else
-		devname = fsprobe_get_devname(spec);
+		devname = spec_to_devname(spec);
 
 	if (devname)
 		mc = getfs_by_devname(devname);
@@ -1841,12 +2112,15 @@ main(int argc, char *argv[]) {
 	initproctitle(argc, argv);
 #endif
 
-	while ((c = getopt_long (argc, argv, "afFhilL:no:O:p:rsU:vVwt:",
+	while ((c = getopt_long (argc, argv, "aBfFhilL:Mno:O:p:rRsU:vVwt:",
 				 longopts, NULL)) != -1) {
 		switch (c) {
 		case 'a':	       /* mount everything in fstab */
 			++mount_all;
 			break;
+		case 'B': /* bind */
+			mounttype = MS_BIND;
+			break;
 		case 'f':	       /* fake: don't actually call mount(2) */
 			++fake;
 			break;
@@ -1865,6 +2139,9 @@ main(int argc, char *argv[]) {
 		case 'L':
 			label = optarg;
 			break;
+		case 'M': /* move */
+			mounttype = MS_MOVE;
+			break;
 		case 'n':		/* do not write /etc/mtab */
 			++nomtab;
 			break;
@@ -1881,6 +2158,9 @@ main(int argc, char *argv[]) {
 			readonly = 1;
 			readwrite = 0;
 			break;
+		case 'R': /* rbind */
+			mounttype = (MS_BIND | MS_REC);
+			break;
 		case 's':		/* allow sloppy mount options */
 			sloppy = 1;
 			break;
@@ -1903,12 +2183,6 @@ main(int argc, char *argv[]) {
 		case 0:
 			break;
 
-		case 128: /* bind */
-			mounttype = MS_BIND;
-			break;
-		case 133: /* move */
-			mounttype = MS_MOVE;
-			break;
 		case 134:
 			/* undocumented, may go away again:
 			   call: mount --guess-fstype device
@@ -1920,9 +2194,6 @@ main(int argc, char *argv[]) {
 			printf("%s\n", fstype ? fstype : "unknown");
 			exit(fstype ? 0 : EX_FAIL);
 		    }
-		case 135:
-			mounttype = (MS_BIND | MS_REC);
-			break;
 
 		case 136:
 			mounttype = MS_SHARED;
@@ -1955,7 +2226,9 @@ main(int argc, char *argv[]) {
 		case 143:
 			mounttype = (MS_UNBINDABLE | MS_REC);
 			break;
-
+		case 144:
+			nocanonicalize = 1;
+			break;
 		case '?':
 		default:
 			usage (stderr, EX_USAGE);
@@ -1967,6 +2240,8 @@ main(int argc, char *argv[]) {
 		printf("mount: mtab path:  \"%s\"\n", _PATH_MOUNTED);
 		printf("mount: lock path:  \"%s\"\n", _PATH_MOUNTED_LOCK);
 		printf("mount: temp path:  \"%s\"\n", _PATH_MOUNTED_TMP);
+		printf("mount: UID:        %d\n", getuid());
+		printf("mount: eUID:       %d\n", geteuid());
 	}
 
 	argc -= optind;
@@ -1980,11 +2255,27 @@ main(int argc, char *argv[]) {
 		return print_all (types);
 	}
 
-	if (getuid () != geteuid ()) {
-		suid = 1;
-		if (types || options || readwrite || nomtab || mount_all ||
-		    fake || mounttype || (argc + specseen) != 1)
+	{
+		const uid_t ruid = getuid();
+		const uid_t euid = geteuid();
+
+		/* if we're really root and aren't running setuid */
+		if (((uid_t)0 == ruid) && (ruid == euid)) {
+			restricted = 0;
+		}
+
+		if (restricted &&
+		    (types || options || readwrite || nomtab || mount_all ||
+		     nocanonicalize || fake || mounttype ||
+		     (argc + specseen) != 1)) {
+
+			if (ruid == 0 && euid != 0)
+				/* user is root, but setuid to non-root */
+				die (EX_USAGE, _("mount: only root can do that "
+					"(effective UID is %d)"), euid);
+
 			die (EX_USAGE, _("mount: only root can do that"));
+		}
 	}
 
 	atexit(unlock_mtab);
